"""Models for responses of api/v1/spin/verify."""

from dataclasses import dataclass, field
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


class VerificationStatus(StrEnum):
    """List of known statuses for SPIN."""

    CORRECT_SPIN = "CORRECT_SPIN"
    INCORRECT_SPIN = "INCORRECT_SPIN"


@dataclass
class SpinStatus(DataClassORJSONMixin):
    state: str
    remaining_tries: int = field(metadata=field_options(alias="remainingTries"))
    locked_waiting_time_in_seconds: int = field(
        metadata=field_options(alias="lockedWaitingTimeInSeconds")
    )


@dataclass
class Spin(DataClassORJSONMixin):
    verification_status: VerificationStatus = field(
        metadata=field_options(alias="verificationStatus")
    )
    spin_status: SpinStatus | None = field(default=None, metadata=field_options(alias="spinStatus"))



"""Models for responses of api/v2/air-conditioning endpoint."""

from dataclasses import dataclass, field
from datetime import datetime, time
from enum import StrEnum
from typing import Any

from mashumaro import field_options
from mashumaro.config import (
    TO_DICT_ADD_BY_ALIAS_FLAG,
    TO_DICT_ADD_OMIT_NONE_FLAG,
    BaseConfig,
)
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .common import ChargerLockedState, ConnectionState, OnOffState, Side, Weekday


class TemperatureUnit(StrEnum):
    CELSIUS = "CELSIUS"


class TimerMode(StrEnum):
    ONE_OFF = "ONE_OFF"
    RECURRING = "RECURRING"


class AirConditioningState(StrEnum):
    COOLING = "COOLING"
    HEATING = "HEATING"
    HEATING_AUXILIARY = "HEATING_AUXILIARY"
    OFF = "OFF"
    ON = "ON"
    VENTILATION = "VENTILATION"
    INVALID = "INVALID"


# Probably other states than AUTOMATIC are available, to be discovered
class HeaterSource(StrEnum):
    AUTOMATIC = "AUTOMATIC"
    ELECTRIC = "ELECTRIC"


@dataclass
class AirConditioningTimer(DataClassORJSONMixin):
    enabled: bool
    id: int
    time: time
    type: TimerMode
    selected_days: list[Weekday] = field(metadata=field_options(alias="selectedDays"))

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        code_generation_options = [  # noqa: RUF012
            TO_DICT_ADD_BY_ALIAS_FLAG
        ]

    def __post_serialize__(self, d: dict[Any, Any]) -> dict[Any, Any]:
        """Post-process the data before serialization."""
        if self.time:
            d["time"] = self.time.strftime("%H:%M")  # Format to hh:mm
        return d


@dataclass
class SeatHeating(DataClassORJSONMixin):
    front_left: bool | None = field(default=None, metadata=field_options(alias="frontLeft"))
    front_right: bool | None = field(default=None, metadata=field_options(alias="frontRight"))

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        code_generation_options = [  # noqa: RUF012
            TO_DICT_ADD_BY_ALIAS_FLAG,
            TO_DICT_ADD_OMIT_NONE_FLAG,
        ]


@dataclass
class TargetTemperature(DataClassORJSONMixin):
    temperature_value: float = field(metadata=field_options(alias="temperatureValue"))
    unit_in_car: TemperatureUnit = field(
        default=TemperatureUnit.CELSIUS, metadata=field_options(alias="unitInCar")
    )

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]  # noqa: RUF012


@dataclass
class OutsideTemperature(TargetTemperature):
    car_captured_timestamp: datetime | None = field(
        default=None, metadata=field_options(alias="carCapturedTimestamp")
    )


@dataclass
class WindowHeatingState(DataClassORJSONMixin):
    front: OnOffState
    rear: OnOffState
    unspecified: Any


@dataclass
class AirConditioningAtUnlock(DataClassORJSONMixin):
    """AirConditioningAtUnlock setting."""

    air_conditioning_at_unlock_enabled: bool = field(
        metadata=field_options(alias="airConditioningAtUnlockEnabled")
    )

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        serialize_by_alias = True


@dataclass
class AirConditioningWithoutExternalPower(DataClassORJSONMixin):
    """AirConditioningWithoutExternalPower setting."""

    air_conditioning_without_external_power_enabled: bool = field(
        metadata=field_options(alias="airConditioningWithoutExternalPowerEnabled")
    )

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        serialize_by_alias = True


@dataclass
class WindowHeating(DataClassORJSONMixin):
    """WindowHeating setting."""

    window_heating_enabled: bool = field(metadata=field_options(alias="windowHeatingEnabled"))

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        serialize_by_alias = True


@dataclass
class AirConditioning(DataClassORJSONMixin):
    """Information related to air conditioning."""

    timers: list[AirConditioningTimer]
    errors: list[Any]
    state: AirConditioningState
    steering_wheel_position: Side | None = field(
        default=None, metadata=field_options(alias="steeringWheelPosition")
    )
    window_heating_state: WindowHeatingState | None = field(
        default=None, metadata=field_options(alias="windowHeatingState")
    )
    car_captured_timestamp: datetime | None = field(
        default=None, metadata=field_options(alias="carCapturedTimestamp")
    )
    air_conditioning_at_unlock: bool | None = field(
        default=None, metadata=field_options(alias="airConditioningAtUnlock")
    )
    charger_connection_state: ConnectionState | None = field(
        default=None, metadata=field_options(alias="chargerConnectionState")
    )
    charger_lock_state: ChargerLockedState | None = field(
        default=None, metadata=field_options(alias="chargerLockState")
    )
    estimated_date_time_to_reach_target_temperature: datetime | None = field(
        default=None, metadata=field_options(alias="estimatedDateTimeToReachTargetTemperature")
    )
    heater_source: HeaterSource | None = field(
        default=None, metadata=field_options(alias="heaterSource")
    )
    seat_heating_activated: SeatHeating | None = field(
        default=None, metadata=field_options(alias="seatHeatingActivated")
    )
    target_temperature: TargetTemperature | None = field(
        default=None, metadata=field_options(alias="targetTemperature")
    )
    window_heating_enabled: bool | None = field(
        default=None, metadata=field_options(alias="windowHeatingEnabled")
    )
    air_conditioning_without_external_power: bool | None = field(
        default=None, metadata=field_options(alias="airConditioningWithoutExternalPower")
    )
    outside_temperature: OutsideTemperature | None = field(
        default=None, metadata=field_options(alias="outsideTemperature")
    )



"""Models for responses of api/v2/vehicle-status/{vin}/driving-range endpoint."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


class EngineType(StrEnum):
    DIESEL = "diesel"
    ELECTRIC = "electric"
    GASOLINE = "gasoline"
    HYBRID = "hybrid"
    CNG = "cng"


@dataclass
class EngineRange(DataClassORJSONMixin):
    engine_type: EngineType = field(metadata=field_options(alias="engineType"))
    current_fuel_level_in_percent: int | None = field(
        default=None, metadata=field_options(alias="currentFuelLevelInPercent")
    )
    current_soc_in_percent: int | None = field(
        default=None, metadata=field_options(alias="currentSoCInPercent")
    )
    remaining_range_in_km: int | None = field(
        default=None, metadata=field_options(alias="remainingRangeInKm")
    )


@dataclass
class DrivingRange(DataClassORJSONMixin):
    car_captured_timestamp: datetime = field(metadata=field_options(alias="carCapturedTimestamp"))
    car_type: EngineType = field(metadata=field_options(alias="carType"))
    primary_engine_range: EngineRange = field(metadata=field_options(alias="primaryEngineRange"))
    secondary_engine_range: EngineRange | None = field(
        default=None, metadata=field_options(alias="secondaryEngineRange")
    )
    total_range_in_km: int | None = field(
        default=None, metadata=field_options(alias="totalRangeInKm")
    )
    ad_blue_range: int | None = field(default=None, metadata=field_options(alias="adBlueRange"))



"""Models for responses of api/v2/vehicle-status/{vin}."""

import logging
from dataclasses import dataclass, field
from datetime import datetime
from enum import IntEnum
from urllib.parse import parse_qs, urlparse

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from myskoda.models.common import DoorLockedState, OnOffState, OpenState

_LOGGER = logging.getLogger(__name__)


class CarBodyElements(IntEnum):
    LEFT_FRONT_DOOR = 0
    RIGHT_FRONT_DOOR = 1
    LEFT_BACK_DOOR = 2
    RIGHT_BACK_DOOR = 3
    LEFT_LIGHT = 4
    RIGHT_LIGHT = 5
    TRUNK = 6
    BONNET = 7
    SUNROOF = 8


class DoorWindowState(IntEnum):
    CLOSED = 1
    WINDOW_OPEN = 2
    DOOR_OPEN = 3
    ALL_OPEN = 4
    UNKNOWN = 0  # default state for invalid values


@dataclass
class Detail(DataClassORJSONMixin):
    bonnet: OpenState
    sunroof: OpenState
    trunk: OpenState


@dataclass
class Overall(DataClassORJSONMixin):
    doors: OpenState
    doors_locked: DoorLockedState = field(metadata=field_options(alias="doorsLocked"))
    lights: OnOffState
    locked: DoorLockedState
    windows: OpenState


@dataclass
class RenderMode(DataClassORJSONMixin):
    one_x: str = field(metadata=field_options(alias="oneX"))
    one_and_half_x: str = field(metadata=field_options(alias="oneAndHalfX"))
    two_x: str = field(metadata=field_options(alias="twoX"))
    three_x: str = field(metadata=field_options(alias="threeX"))


@dataclass
class Renders(DataClassORJSONMixin):
    light_mode: RenderMode = field(metadata=field_options(alias="lightMode"))
    dark_mode: RenderMode = field(metadata=field_options(alias="darkMode"))


@dataclass
class Status(DataClassORJSONMixin):
    """Current status information for a vehicle."""

    detail: Detail
    overall: Overall
    renders: Renders
    car_captured_timestamp: datetime | None = field(
        default=None, metadata=field_options(alias="carCapturedTimestamp")
    )

    def _extract_window_door_state_list_from_url(self) -> list[int]:
        """Extract window/door states from renders url.

        Returns:
           States of doors/windows as a list of integers, other elements states are
           ignored since they are available directly in detail or overall fields.

        """
        try:
            parsed_url = urlparse(self.renders.light_mode.one_x)
            parsed_query = parse_qs(parsed_url.query)
            split_values = parsed_query["vehicleState"][0].split("-")

            # Convert values to integers, replacing invalid ones with default
            integer_map = []
            for value in split_values:
                try:
                    integer_map.append(int(value))
                except ValueError:
                    _LOGGER.warning("Invalid DoorWindowState: '%s', defaulting to UNKNOWN", value)
                    integer_map.append(0)  # Default to UNKNOWN state
        except (IndexError, KeyError, ValueError):
            # Return default if mapping fails
            _LOGGER.exception("Unable to deduct doors/windows state from vehicle status url")
            return [0, 0, 0, 0]
        return integer_map[:4]

    def _get_door_window_state(self, element: CarBodyElements) -> DoorWindowState:
        door_states = self._extract_window_door_state_list_from_url()
        state = door_states[element.value]
        if state in (
            DoorWindowState.CLOSED,
            DoorWindowState.DOOR_OPEN,
            DoorWindowState.WINDOW_OPEN,
            DoorWindowState.ALL_OPEN,
        ):
            return state
        return DoorWindowState.UNKNOWN

    @property
    def left_front_door(self) -> DoorWindowState:
        """State of the left front door."""
        return self._get_door_window_state(CarBodyElements.LEFT_FRONT_DOOR)

    @property
    def right_front_door(self) -> DoorWindowState:
        """State of the right front door."""
        return self._get_door_window_state(CarBodyElements.RIGHT_FRONT_DOOR)

    @property
    def left_back_door(self) -> DoorWindowState:
        """State of the left back door."""
        return self._get_door_window_state(CarBodyElements.LEFT_BACK_DOOR)

    @property
    def right_back_door(self) -> DoorWindowState:
        """State of the right back door."""
        return self._get_door_window_state(CarBodyElements.RIGHT_BACK_DOOR)



"""Models related to service events from the MQTT broker."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum
from typing import Generic, TypeVar

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .charging import ChargeMode, ChargingState, ServiceEventChargingError


class ServiceEventName(StrEnum):
    """List of known Service EventNames."""

    CHANGE_ACCESS = "change-access"
    CHANGE_CHARGE_MODE = "change-charge-mode"
    CHANGE_LIGHTS = "change-lights"
    CHANGE_REMAINING_TIME = "change-remaining-time"
    CHANGE_SOC = "change-soc"
    CHARGING_COMPLETED = "charging-completed"
    CHARGING_ERROR = "charging-error"
    CHARGING_STATUS_CHANGED = "charging-status-changed"
    CLIMATISATION_COMPLETED = "climatisation-completed"
    DEPARTURE_READY = "departure-ready"
    DEPARTURE_STATUS_CHANGED = "departure-status-changed"


@dataclass
class ServiceEventData(DataClassORJSONMixin):
    """Base class for data in service events."""

    user_id: str = field(metadata=field_options(alias="userId"))
    vin: str


T = TypeVar("T", bound=ServiceEventData)


@dataclass
class ServiceEvent(Generic[T], DataClassORJSONMixin):
    """Main model for Service Events.

    Service Events are unsolicited events emitted by the MQTT bus towards the client.
    """

    version: int
    producer: str
    name: ServiceEventName
    data: T
    trace_id: str = field(metadata=field_options(alias="traceId"))
    timestamp: datetime | None = field(default=None)


def _deserialize_mode(value: str) -> ChargeMode:  # noqa: PLR0911
    match value:
        case "homeStorageCharging":
            return ChargeMode.HOME_STORAGE_CHARGING
        case "immediateDischarging":
            return ChargeMode.IMMEDIATE_DISCHARGING
        case "onlyOwnCurrent":
            return ChargeMode.ONLY_OWN_CURRENT
        case "preferredChargingTimes":
            return ChargeMode.PREFERRED_CHARGING_TIMES
        case "timerChargingWithClimatisation":
            return ChargeMode.TIMER_CHARGING_WITH_CLIMATISATION
        case "timer":
            return ChargeMode.TIMER
        case "manual":
            return ChargeMode.MANUAL
        case "off":
            return ChargeMode.OFF
        case _:
            raise UnexpectedChargeModeError


def _deserialize_charging_state(value: str) -> ChargingState:
    match value:
        case "charging":
            return ChargingState.CHARGING
        case "chargePurposeReachedAndConservation":
            return ChargingState.CONSERVING
        case "chargePurposeReachedAndNotConservationCharging":
            return ChargingState.READY_FOR_CHARGING
        case "notReadyForCharging":
            return ChargingState.CONNECT_CABLE
        case "readyForCharging":
            return ChargingState.READY_FOR_CHARGING
        case "conserving":
            return ChargingState.CONSERVING
        case _:
            raise UnexpectedChargingStateError


def _deserialize_time_to_finish(value: int | str) -> int | None:
    if value == "null":
        return None
    return int(value)


@dataclass
class ServiceEventChargingData(ServiceEventData):
    """Charging data inside charging service event change-soc."""

    mode: ChargeMode | None = field(
        default=None, metadata=field_options(deserialize=_deserialize_mode)
    )
    state: ChargingState | None = field(
        default=None, metadata=field_options(deserialize=_deserialize_charging_state)
    )
    soc: int | None = field(default=None)
    charged_range: int | None = field(default=None, metadata=field_options(alias="chargedRange"))
    time_to_finish: int | None = field(
        default=None,
        metadata=field_options(alias="timeToFinish", deserialize=_deserialize_time_to_finish),
    )
    error_code: ServiceEventChargingError | None = field(
        default=None, metadata=field_options(alias="errorCode")
    )


@dataclass
class ServiceEventWithChargingData(ServiceEvent):
    data: ServiceEventChargingData


class UnexpectedChargeModeError(Exception):
    pass


class UnexpectedChargingStateError(Exception):
    pass



"""Events emitted by MySkoda."""

from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from typing import Literal

from mashumaro.mixins.orjson import DataClassORJSONMixin

from .models.operation_request import OperationRequest
from .models.service_event import ServiceEvent


class ServiceEventTopic(StrEnum):
    ACCESS = "ACCESS"
    AIR_CONDITIONING = "AIR_CONDITIONING"
    AUXILIARY_HEATING = "AUXILIARY_HEATING"
    CHARGING = "CHARGING"
    LIGHTS = "LIGHTS"
    DEPARTURE = "DEPARTURE"


class AccountEventTopic(StrEnum):
    ACCOUNT_PRIVACY = "ACCOUNT_PRIVACY"


class EventType(StrEnum):
    ACCOUNT_EVENT = "account-event"
    OPERATION = "operation-request"
    SERVICE_EVENT = "service-event"


@dataclass
class BaseEvent(DataClassORJSONMixin):
    vin: str
    user_id: str
    timestamp: datetime


@dataclass
class EventOperation(BaseEvent):
    operation: OperationRequest
    type: Literal[EventType.OPERATION] = EventType.OPERATION


@dataclass
class EventAirConditioning(BaseEvent):
    event: ServiceEvent
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.AIR_CONDITIONING] = ServiceEventTopic.AIR_CONDITIONING


@dataclass
class EventAuxiliaryHeating(BaseEvent):
    event: ServiceEvent
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.AUXILIARY_HEATING] = ServiceEventTopic.AUXILIARY_HEATING


@dataclass
class EventCharging(BaseEvent):
    event: ServiceEvent
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.CHARGING] = ServiceEventTopic.CHARGING


@dataclass
class EventAccess(BaseEvent):
    event: ServiceEvent
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.ACCESS] = ServiceEventTopic.ACCESS


@dataclass
class EventLights(BaseEvent):
    event: ServiceEvent
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.LIGHTS] = ServiceEventTopic.LIGHTS


@dataclass
class EventDeparture(BaseEvent):
    event: ServiceEvent
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.DEPARTURE] = ServiceEventTopic.DEPARTURE


@dataclass
class EventAccountPrivacy(BaseEvent):
    type: Literal[EventType.ACCOUNT_EVENT] = EventType.ACCOUNT_EVENT
    topic: Literal[AccountEventTopic.ACCOUNT_PRIVACY] = AccountEventTopic.ACCOUNT_PRIVACY


Event = (
    EventAccountPrivacy
    | EventOperation
    | EventAccess
    | EventAirConditioning
    | EventAuxiliaryHeating
    | EventCharging
    | EventLights
    | EventDeparture
)



"""Models for responses of api/v1/vehicle-automatization/{vin}/departure/ endpoint."""

from dataclasses import dataclass, field
from datetime import datetime
from datetime import time as datetime_time
from typing import Any

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .air_conditioning import TemperatureUnit, TimerMode
from .common import Weekday


@dataclass
class DepartureTemperature(DataClassORJSONMixin):
    celsius: float | None = field(default=None, metadata=field_options(alias="celsius"))
    fahrenheit: float | None = field(default=None, metadata=field_options(alias="fahrenheit"))
    unit_in_car: TemperatureUnit = field(
        default=TemperatureUnit.CELSIUS, metadata=field_options(alias="unitInCar")
    )


@dataclass
class ChargingTime(DataClassORJSONMixin):
    """Information related to DepartureTimer."""

    id: int
    enabled: bool
    start_time: datetime_time = field(metadata=field_options(alias="startTime"))
    end_time: datetime_time = field(metadata=field_options(alias="endTime"))

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        serialize_by_alias = True

    def __post_serialize__(self, d: dict[Any, Any]) -> dict[Any, Any]:
        """Post-process the data before serialization."""
        if self.start_time:
            d["startTime"] = self.start_time.strftime("%H:%M")  # Format to hh:mm
        if self.end_time:
            d["endTime"] = self.end_time.strftime("%H:%M")  # Format to hh:mm
        return d


@dataclass
class DepartureTimer(DataClassORJSONMixin):
    """Information related to DepartureTimer."""

    id: int
    enabled: bool
    charging: bool | None = field(default=None, metadata=field_options(alias="charging"))
    climatisation: bool | None = field(default=None, metadata=field_options(alias="climatisation"))
    preferred_charging_times: list[ChargingTime] | None = field(
        default=None, metadata=field_options(alias="preferredChargingTimes")
    )
    one_off_day: Weekday | None = field(default=None, metadata=field_options(alias="oneOffDay"))
    recurring_on: list[Weekday] | None = field(
        default=None, metadata=field_options(alias="recurringOn")
    )
    target_battery_state_of_charge_in_percent: int | None = field(
        default=None, metadata=field_options(alias="targetBatteryStateOfChargeInPercent")
    )
    time: datetime_time | None = field(default=None, metadata=field_options(alias="time"))
    type: TimerMode | None = field(default=None, metadata=field_options(alias="type"))

    class Config(BaseConfig):
        """Configuration for serialization and deserialization.."""

        serialize_by_alias = True
        omit_none = True

    def __post_serialize__(self, d: dict[Any, Any]) -> dict[Any, Any]:
        """Post-process the data before serialization."""
        if self.time:
            d["time"] = self.time.strftime("%H:%M")  # Format to hh:mm
        return d


@dataclass
class DepartureSettings(DataClassORJSONMixin):
    """Information related to DepartureSettings."""

    minimum_battery_state_of_charge_in_percent: int | None = field(
        default=None, metadata=field_options(alias="minimumBatteryStateOfChargeInPercent")
    )


@dataclass
class DepartureInfo(DataClassORJSONMixin):
    """Information related to Departure."""

    car_captured_timestamp: datetime | None = field(
        default=None, metadata=field_options(alias="carCapturedTimestamp")
    )
    first_occurring_timer_id: int | None = field(
        default=None, metadata=field_options(alias="firstOccurringTimerId")
    )
    minimum_battery_state_of_charge_in_percent: int | None = field(
        default=None, metadata=field_options(alias="minimumBatteryStateOfChargeInPercent")
    )
    target_temperature: DepartureTemperature | None = field(
        default=None, metadata=field_options(alias="targetTemperature")
    )
    timers: list[DepartureTimer] | None = field(
        default=None, metadata=field_options(alias="timers")
    )
